"""get_file_list module gets the file list from a path for both readers"""

import fsspec
from typing import List, Tuple, Union
import os


def get_file_list(path: Union[str, List[str]], file_format: str) -> Tuple[fsspec.AbstractFileSystem, List[str]]:
    """
    Get the file system and all the file paths that matches `file_format` under the given `path`.
    The `path` could a single folder or multiple folders.
    :raises ValueError: if file system is inconsistent under different folders.
    """
    if isinstance(path, str):
        return _get_file_list(path, file_format)
    all_file_paths = []
    fs = None
    for p in path:
        cur_fs, file_paths = _get_file_list(p, file_format, sort_result=False)
        if fs is None:
            fs = cur_fs
        elif fs != cur_fs:
            raise ValueError(
                f"The file system in different folder are inconsistent.\n" f"Got one {fs} and the other {cur_fs}"
            )
        all_file_paths.extend(file_paths)
    all_file_paths.sort()
    return fs, all_file_paths


def make_path_absolute(path: str) -> str:
    fs, p = fsspec.core.url_to_fs(path)
    if fs.protocol == "file":
        return os.path.abspath(p)
    return path


def _get_file_list(
    path: str, file_format: str, sort_result: bool = True
) -> Tuple[fsspec.AbstractFileSystem, List[str]]:
    """Get the file system and all the file paths that matches `file_format` given a single path."""
    path = make_path_absolute(path)
    fs, path_in_fs = fsspec.core.url_to_fs(path)
    prefix = path[: path.index(path_in_fs)]
    glob_pattern = path.rstrip("/") + f"**/*.{file_format}"
    file_paths = fs.glob(glob_pattern)
    if sort_result:
        file_paths.sort()
    file_paths_with_prefix = [prefix + file_path for file_path in file_paths]
    return fs, file_paths_with_prefix



"""Parquet embedding reader, read embeddings from parquet files in streaming"""

import pandas as pd
from multiprocessing.pool import ThreadPool
from tqdm import tqdm
import numpy as np
import pyarrow.parquet as pq
from collections import namedtuple
from embedding_reader.get_file_list import get_file_list
from embedding_reader.piece_builder import build_pieces, PIECES_BASE_COLUMNS
from threading import Semaphore
import math


class ParquetReader:
    """Parquet reader class, implements init to read the files headers and call to produce embeddings batches"""

    def __init__(self, embeddings_folder, embedding_column_name, metadata_column_names=None):
        self.embeddings_folder = embeddings_folder
        self.fs, embeddings_file_paths = get_file_list(embeddings_folder, "parquet")

        self.metadata_column_names = metadata_column_names
        self.embedding_column_name = embedding_column_name

        # read one non empty file to get the dimension
        for filename in embeddings_file_paths:
            with self.fs.open(filename, "rb") as f:
                parquet_file = pq.ParquetFile(f, memory_map=True)
                batches = parquet_file.iter_batches(batch_size=1, columns=[embedding_column_name])
                try:
                    embedding = next(batches).to_pandas()[embedding_column_name].to_numpy()[0]
                    self.dimension = int(embedding.shape[0])
                    break
                except StopIteration:
                    continue

        def file_to_header(filename):
            try:
                with self.fs.open(filename, "rb") as f:
                    parquet_file = pq.ParquetFile(f, memory_map=True)
                    return (None, [filename, parquet_file.metadata.num_rows])
            except Exception as e:  # pylint: disable=broad-except
                return e, (filename, None)

        headers = []
        count_before = 0
        with ThreadPool(10) as p:
            for err, c in tqdm(p.imap(file_to_header, embeddings_file_paths), total=len(embeddings_file_paths)):
                if err is not None:
                    raise Exception(f"failed reading file {c[0]}") from err
                if c[1] == 0:
                    continue
                headers.append([*c, count_before])
                count_before += c[1]

        self.headers = pd.DataFrame(headers, columns=["filename", "count", "count_before"])
        self.count = self.headers["count"].sum()
        if self.count == 0:
            raise ValueError("No embeddings found in folder {}".format(embeddings_folder))
        self.byte_per_item = 4 * self.dimension

        self.total_size = self.count * self.byte_per_item

    def __call__(self, batch_size, start=0, end=None, max_piece_size=None, parallel_pieces=None, show_progress=True):
        if end is None:
            end = self.count

        if end > self.count:
            end = self.count
        if batch_size > end - start:
            batch_size = end - start

        if max_piece_size is None:
            max_piece_size = max(int(50 * 10 ** 6 / (self.byte_per_item)), 1)
        if parallel_pieces is None:
            parallel_pieces = max(math.ceil(batch_size / max_piece_size), 10)

        pieces = build_pieces(
            headers=self.headers, batch_size=batch_size, start=start, end=end, max_piece_size=max_piece_size
        )

        cols = PIECES_BASE_COLUMNS
        Piece = namedtuple("Count", cols)

        def read_piece(piece):
            try:
                start = piece.piece_start
                end = piece.piece_end
                path = piece.filename

                with self.fs.open(path, "rb") as f:
                    length = end - start
                    table = pq.read_table(f)
                    id_columns = self.metadata_column_names
                    table_slice = table.slice(start, length)
                    embeddings_raw = table_slice[self.embedding_column_name].to_numpy()
                    ids = table_slice.select(id_columns).to_pandas() if id_columns else None

                    return (None, (np.stack(embeddings_raw), ids, piece,))
            except Exception as e:  # pylint: disable=broad-except
                return e, (None, None, piece)

        semaphore = Semaphore(parallel_pieces)

        def piece_generator(pieces):
            for piece in (Piece(*parts) for parts in zip(*[pieces[col] for col in cols])):
                semaphore.acquire()
                yield piece

        batch = None
        batch_meta = None
        batch_offset = 0

        if show_progress:
            pbar = tqdm(total=len(pieces))
        with ThreadPool(parallel_pieces) as p:
            for err, (data, meta, piece) in p.imap(read_piece, piece_generator(pieces)):
                if err is not None:
                    semaphore.release()
                    raise Exception(
                        f"failed reading file {piece.filename} from {piece.piece_start} to {piece.piece_end}"
                    ) from err
                if batch is None:
                    batch = np.empty((piece.batch_length, self.dimension), "float32")
                    if self.metadata_column_names is not None:
                        batch_meta = np.empty((piece.batch_length, len(self.metadata_column_names) + 1), dtype="object")
                batch[batch_offset : (batch_offset + piece.piece_length)] = data
                if self.metadata_column_names is not None:
                    batch_meta[batch_offset : (batch_offset + piece.piece_length)] = meta.to_numpy()
                batch_offset += data.shape[0]
                if piece.last_piece:
                    if self.metadata_column_names is not None:
                        meta_batch_df = pd.DataFrame(
                            batch_meta, columns=self.metadata_column_names + ["i"]
                        ).infer_objects()
                        meta_batch_df["i"] = np.arange(start=piece.batch_start, stop=piece.batch_end)
                    else:
                        meta_batch_df = None
                    yield batch, meta_batch_df
                    batch = None
                    if self.metadata_column_names is not None:
                        batch_meta = None
                    batch_offset = 0

                if show_progress:
                    pbar.update(1)
                semaphore.release()

        if show_progress:
            pbar.close()



"""Numpy embedding reader, read embeddings from numpy files in streaming

The main logic of this reader is:
* read file headers to know the length and dimensions
* compute pieces to read from each file depending on batch size and max piece length
* read pieces in parallel
* concatenate pieces
"""

import pandas as pd
from multiprocessing.pool import ThreadPool
from tqdm import tqdm
import numpy as np
import re
import math
from collections import namedtuple
from embedding_reader.get_file_list import get_file_list
from embedding_reader.piece_builder import build_pieces, PIECES_BASE_COLUMNS
from threading import Semaphore


def read_numpy_header(f):
    """Read the header of a numpy file"""
    f.seek(0)
    file_size = f.size if isinstance(f.size, int) else f.size()
    first_line = f.read(min(file_size, 300)).split(b"\n")[0]
    result = re.search(r"'shape': \(([0-9]+), ([0-9]+)\)", str(first_line))
    shape = (int(result.group(1)), int(result.group(2)))
    dtype = re.search(r"'descr': '([<f0-9]+)'", str(first_line)).group(1)
    end = len(first_line) + 1  # the first line content and the endline
    f.seek(0)
    byte_per_item = np.dtype(dtype).itemsize * shape[1]
    return (shape[0], shape[1], dtype, end, byte_per_item)


class NumpyReader:
    """Numpy reader class, implements init to read the files headers and call to procuce embeddings batches"""

    def __init__(self, embeddings_folder):
        self.embeddings_folder = embeddings_folder
        self.fs, embeddings_file_paths = get_file_list(embeddings_folder, "npy")

        def file_to_header(filename):
            try:
                with self.fs.open(filename, "rb") as f:
                    return (None, [filename, *read_numpy_header(f)])
            except Exception as e:  # pylint: disable=broad-except
                return e, (filename, None)

        headers = []
        count_before = 0
        with ThreadPool(10) as p:
            for err, c in tqdm(p.imap(file_to_header, embeddings_file_paths), total=len(embeddings_file_paths)):
                if err is not None:
                    raise Exception(f"failed reading file {c[0]}") from err
                if c[1] == 0:
                    continue
                headers.append([*c[0:2], count_before, *c[2:]])
                count_before += c[1]

        self.headers = pd.DataFrame(
            headers,
            columns=["filename", "count", "count_before", "dimension", "dtype", "header_offset", "byte_per_item"],
        )

        self.count = self.headers["count"].sum()
        if self.count == 0:
            raise ValueError("No embeddings found in folder {}".format(embeddings_folder))
        self.dimension = int(self.headers.iloc[0]["dimension"])
        self.byte_per_item = self.headers.iloc[0]["byte_per_item"]
        self.dtype = self.headers.iloc[0]["dtype"]
        self.total_size = self.count * self.byte_per_item

    def __call__(self, batch_size, start=0, end=None, max_piece_size=None, parallel_pieces=None, show_progress=True):
        if end is None:
            end = self.count

        if end > self.count:
            end = self.count
        if batch_size > end - start:
            batch_size = end - start

        if max_piece_size is None:
            max_piece_size = max(int(50 * 10 ** 6 / (self.byte_per_item)), 1)
        if parallel_pieces is None:
            parallel_pieces = max(math.ceil(batch_size / max_piece_size), 10)

        metadata_columns = ["header_offset"]
        pieces = build_pieces(
            headers=self.headers,
            batch_size=batch_size,
            start=start,
            end=end,
            max_piece_size=max_piece_size,
            metadata_columns=metadata_columns,
        )
        cols = PIECES_BASE_COLUMNS + metadata_columns
        Piece = namedtuple("Count", cols)

        def read_piece(piece):
            try:
                start = piece.piece_start
                end = piece.piece_end
                path = piece.filename
                header_offset = piece.header_offset

                with self.fs.open(path, "rb") as f:
                    length = end - start
                    f.seek(header_offset + start * self.byte_per_item)
                    return (
                        None,
                        (
                            np.frombuffer(f.read(length * self.byte_per_item), dtype=self.dtype).reshape(
                                (length, self.dimension)
                            ),
                            piece,
                        ),
                    )
            except Exception as e:  # pylint: disable=broad-except
                return e, (None, piece)

        semaphore = Semaphore(parallel_pieces)

        def piece_generator(pieces):
            for piece in (Piece(*parts) for parts in zip(*[pieces[col] for col in cols])):
                semaphore.acquire()
                yield piece

        batch = None
        batch_offset = 0

        if show_progress:
            pbar = tqdm(total=len(pieces))
        with ThreadPool(parallel_pieces) as p:
            for err, (data, piece) in p.imap(read_piece, piece_generator(pieces)):
                if err is not None:
                    semaphore.release()
                    raise Exception(
                        f"failed reading file {piece.filename} from {piece.piece_start} to {piece.piece_end}"
                    ) from err

                if batch is None:
                    batch = np.empty((piece.batch_length, self.dimension), "float32")

                batch[batch_offset : (batch_offset + piece.piece_length)] = data
                batch_offset += data.shape[0]
                if piece.last_piece:
                    yield batch, None
                    batch = None
                    batch_offset = 0

                if show_progress:
                    pbar.update(1)
                semaphore.release()

        if show_progress:
            pbar.close()



"""Reader module exposes the reading functionality of all formats"""

from embedding_reader.numpy_reader import NumpyReader
from embedding_reader.parquet_reader import ParquetReader


class EmbeddingReader:
    """reader class, implements init to read the files headers and call to produce embeddings batches"""

    def __init__(self, embeddings_folder, file_format="npy", embedding_column="embedding", meta_columns=None):
        if file_format == "npy":
            self.reader = NumpyReader(embeddings_folder)
        elif file_format == "parquet":
            self.reader = ParquetReader(
                embeddings_folder, embedding_column_name=embedding_column, metadata_column_names=meta_columns
            )
        else:
            raise ValueError("format must be npy or parquet")

        self.dimension = self.reader.dimension
        self.count = self.reader.count
        self.byte_per_item = self.reader.byte_per_item
        self.total_size = self.reader.total_size
        self.embeddings_folder = embeddings_folder

    def __call__(self, batch_size, start=0, end=None, max_piece_size=None, parallel_pieces=10, show_progress=True):
        return self.reader(batch_size, start, end, max_piece_size, parallel_pieces, show_progress)



"""piece builder module provides the build_pieces function"""

import pandas as pd
from collections import namedtuple


PIECES_BASE_COLUMNS = [
    "filename",
    "piece_start",
    "piece_end",
    "piece_length",
    "batch_id",
    "batch_start",
    "batch_end",
    "batch_length",
    "last_piece",
]


def build_pieces(headers, batch_size, start, end, max_piece_size=100000, metadata_columns=None):
    """
    Build pieces function takes as input a list of headers and
    returns a list of pieces split in size maximum max_piece_size.
    Input: (filename, count, count_before)
    Output: (filename:str, piece_start:int, piece_end:int, batch_id:int, batch_length:int, last_piece:bool)

    This function is the main feature of embedding reader, it makes it possible to read many files
    in parallel and abstract away all the batching
    """

    if metadata_columns is None:
        metadata_columns = []

    columns = ["filename", "count", "count_before"] + metadata_columns

    filecount = namedtuple("FileCount", columns)

    items = [filecount(*args) for args in zip(*[headers[col] for col in columns])]

    header_i = 0
    while header_i < len(items) and items[header_i].count_before + items[header_i].count <= start:
        header_i += 1
        continue

    if header_i == len(items):
        raise ValueError(f"Can not build pieces for batch with start: {start}, end: {end}, perhaps reduce the start")

    # we skipped start-count_before from the first file
    read_current_file = start - items[header_i].count_before
    read_current_batch = 0
    pieces = []

    for batch_id, batch_start in enumerate(range(start, end, batch_size)):
        batch_length = min(batch_size, end - batch_start)

        # building all pieces of this batch
        while header_i < len(items) and read_current_batch < batch_length:
            if items[header_i].count == 0:
                header_i += 1
                continue
            count_before = items[header_i].count_before
            count = items[header_i].count
            piece_start = batch_start + read_current_batch - count_before
            piece_length = min(count - piece_start, batch_length - read_current_batch, max_piece_size)
            piece_end = piece_start + piece_length

            read_current_file += piece_length
            read_current_batch += piece_length
            piece_filename = items[header_i].filename
            last_piece = read_current_batch == batch_length
            batch_end = batch_start + batch_length
            piece = (
                piece_filename,
                piece_start,
                piece_end,
                piece_length,
                batch_id,
                batch_start,
                batch_end,
                batch_length,
                last_piece,
            )
            piece = piece + tuple(items[header_i][3 + col] for col in range(len(metadata_columns)))

            pieces.append(piece)

            if read_current_file == items[header_i].count:
                read_current_file = 0
                header_i += 1

        read_current_batch = 0

    return pd.DataFrame(pieces, columns=PIECES_BASE_COLUMNS + metadata_columns,)
